/*	
**	
**	Copyright (C) 1999 Geert Bevin
**	gbevin@thunderstorms.org
**	http://thunderstorms.org/gclipper
**
**	This program is free software; you can redistribute it and/or modify
**	it under the terms of the GNU General Public License as published by
**	the Free Software Foundation; either version 2 of the License, or
**	(at your option) any later version.
**	
**	This program is distributed in the hope that it will be useful,
**	but WITHOUT ANY WARRANTY; without even the implied warranty of
**	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**	GNU General Public License for more details.
**	
**	You should have received a copy of the GNU General Public License
**	along with this program; if not, write to the Free Software
**	
**	Foundation,Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
**		
*/

#include <X11/Xlib.h>
#include <X11/Xatom.h>
#include <stdio.h>
#include <string.h>
#include <gdk/gdk.h>
#include <gdk/gdkprivate.h>
#include <gdk/gdkx.h>
#include <gtk/gtk.h>

#define NO_TARGETCOPY  1
#define MAX_SELECTIONS 20

GdkAtom SELECTION_PRIMARY = GDK_SELECTION_PRIMARY;
GdkAtom SELECTION_CLIPBOARD;

/************************************************************************
 * Structure to store the type of the target, the length of the data    *
 * and a pointer to the data that has been copied.                      *
 ************************************************************************/
struct TargetCopy
{
    gchar   *name;
    GdkAtom  type;
    gint     format;
    gpointer data;
    gint     length;
};

/************************************************************************
 * Set the owner button on.                                             *
 ************************************************************************/
void set_owner_button_on(GtkWindow *window)
{
    GtkWidget *button_owner = NULL;
    
    button_owner = gtk_object_get_data(GTK_OBJECT(window),
        "button_owner");
        
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(button_owner), TRUE);
}

/************************************************************************
 * Set the owner button off.                                            *
 ************************************************************************/
void set_owner_button_off(GtkWindow *window)
{
    GtkWidget *button_owner = NULL;

    button_owner = gtk_object_get_data(GTK_OBJECT(window),
        "button_owner");

    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(button_owner), FALSE);
}

/************************************************************************
 * Take the ownership of the primary selection.                         *
 ************************************************************************/
void become_primary_selection_owner(GtkWindow *window, GtkWidget *widget)
{
    set_owner_button_on(window);

    /* Take the selection ownership. */
    gtk_selection_owner_set(widget, SELECTION_PRIMARY, GDK_CURRENT_TIME);
}

/************************************************************************
 * Take the ownership of the clipboard selection.                       *
 ************************************************************************/
void become_clipboard_selection_owner(GtkWindow *window,
     GtkWidget *widget)
{
    /* Take the selection ownership. */
    gtk_selection_owner_set(widget, SELECTION_CLIPBOARD,
        GDK_CURRENT_TIME);
}

/************************************************************************
 * Give up the ownership of a selection.                                *
 ************************************************************************/
void give_up_selection_ownership(GtkWindow *window, GdkAtom selection)
{
    set_owner_button_off(window);

    /* Give up the selection ownership. */
    gtk_selection_owner_set(NULL, selection, GDK_CURRENT_TIME);
}

/************************************************************************
 * Send an owner widget the request to receive the data for a certain   *
 * target                                                               *
 ************************************************************************/
void get_target(GtkWidget *widget, gchar *target, GdkAtom selection)
{
    GdkAtom  atom_target = GDK_NONE;
    gboolean convert_successfull = FALSE;
    
    /* Get the atom that corresponds to the target name. */
    atom_target = gdk_atom_intern(target, FALSE); 

    do
    {
        /* Ask the owner to convert the selection to the target format,
           if another request is already being processed by the owner,
           the function returns FALSE. If the conversion was sucessfull,
           TRUE is returned. */
        convert_successfull = gtk_selection_convert(widget, selection,
            atom_target, GDK_CURRENT_TIME);
        
        /* Let Gtk process all the pending events before continuing. */
        while(gtk_events_pending())
        {
            gtk_main_iteration();
        }
    }
    /* Repeat until the conversion was successfull */
    while(!convert_successfull);
}

/************************************************************************
 * Get the targets and data of a specific selection. The data will be   *
 * recieved by the window_selection_received() callback that is called  *
 * when the main window get the selection_receive signal.               *
 ************************************************************************/
void selection_get(GtkWidget *window, GdkAtom selection)
{
    /* Get the TIMESTAMP target of the selection. The TARGETS target is 
       not requested first to provide a suitable delay for selections that
       or performed by double-clicking some text. Otherwise, the selection
       made with the first click would also be fetched. */
    get_target(window, "TIMESTAMP", selection);
}

/************************************************************************
 * Get the selection data 'manually' which means that it's not an       *
 * automatically triggered event, but that it has been somehow          *
 * requested by the user.                                               *
 ************************************************************************/
void selection_get_manual(GtkWidget *window, GdkAtom selection)
{
    /* Set the main window object to flag the the selection retrieval as
       being manually triggered. */
    gtk_object_set_data(GTK_OBJECT(window), "MANUAL_GET",
        (gpointer)1);
        
    selection_get(window, selection);
}

/************************************************************************
 * Function to be able to get the primary selection by providing it as  *
 * argument to a gtk_timeout_add() function.                            *
 ************************************************************************/
gint selection_get_primary(GtkWidget *window)
{
    selection_get(window, SELECTION_PRIMARY);
    
    return FALSE;
}

/************************************************************************
 * Function to be able to get the clipboard selection by providing it   *
 * as argument to a gtk_timeout_add() function.                         *
 ************************************************************************/
gint selection_get_clipboard(GtkWidget *window)
{
    selection_get(window, SELECTION_CLIPBOARD);
    
    return FALSE;
}

/************************************************************************
 * Check if the owner of a particular selection is part of gclipper.    *
 ************************************************************************/
gboolean is_owner_gclipper(GtkWindow *window, GdkAtom atom_selection)
{
    GList     *clist_notebook_windows = NULL;
    GdkWindow *owner_window = NULL;

    clist_notebook_windows = gtk_object_get_data(GTK_OBJECT(window),
        "clist_notebook_windows");
        
    /* Get the gdk window of the new selection owner, returns NULL when
       the owner is not a gdk window but another kind of X window. */
    owner_window = gdk_selection_owner_get(atom_selection);

    /* Check if the new owner is a member of the list of notebook clist
       gdk windows. If this is not the case, return the result that the
       current selection is not owned by a child widget of the gclipper
       application. */    
    if(g_list_find(clist_notebook_windows, owner_window))
    {
        return TRUE;
    }
    else
    {
        return FALSE;
    }
}

/************************************************************************
 * Synchronize the state of the delete button, details button, owner    *
 * button and notebook with the contents of the main clist.             *
 ************************************************************************/
void synchronize_state(GtkWindow *window)
{
    GtkWidget *clist_main = NULL;
    GtkWidget *notebook = NULL;
    GtkWidget *button_delete = NULL;
    GtkWidget *button_details = NULL;

    clist_main = gtk_object_get_data(GTK_OBJECT(window), "clist");
    notebook = gtk_object_get_data(GTK_OBJECT(window), "notebook");
    button_delete = gtk_object_get_data(GTK_OBJECT(window),
        "button_delete");
    button_details = gtk_object_get_data(GTK_OBJECT(window),
        "button_details");

    /* Check if there are rows present in the main clist. */
    if(GTK_CLIST(clist_main)->rows)
    {
        gtk_widget_set_sensitive(button_delete, TRUE);
        gtk_widget_set_sensitive(button_details, TRUE);
    }
    else
    {
        gtk_widget_set_sensitive(button_delete, FALSE);
        gtk_widget_set_sensitive(button_details, FALSE);
        gtk_toggle_button_set_state(GTK_TOGGLE_BUTTON(button_details),
            FALSE);
        gtk_widget_hide(notebook);
    }
}

/************************************************************************
 * Signal handler that is called when a row in the main clist has       *
 * been selected.                                                       *
 ************************************************************************/
void clist_main_select_row(GtkCList *clist, gint row, gint column,
         GdkEventButton *event, GtkWidget *window)
{
    GtkWidget   *notebook = NULL;
    GtkWidget   *notebook_page = NULL;
    GtkWidget   *clist_notebook = NULL;
    GtkTooltips *tooltips = NULL;

    notebook = gtk_object_get_data(GTK_OBJECT(window), "notebook");
    tooltips = gtk_object_get_data(GTK_OBJECT(window), "tooltips");
    
    /* Change the visible notebook page. */
    gtk_notebook_set_page(GTK_NOTEBOOK(notebook), row);
    
    notebook_page = gtk_notebook_get_nth_page(GTK_NOTEBOOK(notebook),
        row);
    clist_notebook = gtk_object_get_data(GTK_OBJECT(notebook_page),
        "clist");
        
    /* Make the notebook page clist widget the new selection owner for
       both the primary and clipboard selection. */
    become_primary_selection_owner(GTK_WINDOW(window), clist_notebook);
    become_clipboard_selection_owner(GTK_WINDOW(window), clist_notebook);
}

/************************************************************************
 * Signal handler that is called when a mouse button has been clicked   *
 * the main clist.                                                      *
 ************************************************************************/
void clist_main_button_press(GtkCList *clist, GdkEventButton *event,
         GtkWindow *window)
{
    /* If the 2nd mouse button is pressed, get the selection.
       This provides a similar behaviour as when you press the 2nd
       mousebutton in other applications (mostly text based). */
    if(event->button == 2)
    {
        selection_get_manual(GTK_WIDGET(window), SELECTION_PRIMARY);
    }
}

/************************************************************************
 * Signal handler that is called when the owner of the selection        *
 * notifies the notebook clist that conversion has been completed and   *
 * that the selection data has been provided.                           *
 ************************************************************************/
void clist_notebook_selection_received(GtkWidget *clist_notebook,
         GtkSelectionData *data, guint time, GtkWidget *window)
{
    gchar              *atom_targetname = NULL;
    gint               row = 0;
    struct TargetCopy  *target_copy = NULL;
    
    /* Get the string representation of the atom that defines the target
       that was received. */
    atom_targetname = gdk_atom_name(data->target);
    
    /* Get the row number that is associated with the target name by
       looking it up in the notebook clist object data. */
    row = (gint)gtk_object_get_data(GTK_OBJECT(clist_notebook),
        atom_targetname) - 1;
    
    /* Make a copy of the target structure by copying the name, type,
       length, format and data. */
    target_copy = g_new0(struct TargetCopy, 1);
    target_copy->name = g_strdup(atom_targetname);
    target_copy->type = data->type;
    target_copy->length = data->length;
    target_copy->format = data->format;
    if(data->length == 0)
    {
        target_copy->data = NULL;
    }
    else
    {
        target_copy->data = g_memdup(data->data, data->length+1);
    }

    /* Free the name of the atom. */
    g_free(atom_targetname);
    
    /* Register the target both for the primary and the clipboard
       selection. This performs then a kind of selection 'conversion'.
       Fetched primary selections become useable as clipboard selections
       and vice-versa. */
    gtk_selection_add_target(GTK_WIDGET(clist_notebook),
        SELECTION_PRIMARY, data->target, 0);
    gtk_selection_add_target(GTK_WIDGET(clist_notebook),
        SELECTION_CLIPBOARD, data->target, 0);
    
    /* Store a pointer to the target copy as data that is attached to the
       corresponding notebook clist row. */
    gtk_clist_set_row_data(GTK_CLIST(clist_notebook),
        row, target_copy);
}

/************************************************************************
 * Signal handler that is called when a selection requestor asks for    *
 * the selection in a specific target format.                           *
 ************************************************************************/
void clist_notebook_selection_get(GtkCList *clist_notebook,
         GtkSelectionData *data, guint info, guint time,
         GtkWidget *window)
{
    gchar             *target_name = NULL;
    gint              target_row = 0;
    struct TargetCopy *target_copy = NULL;
    
    /* Get the name of the requested target. */
    target_name = gdk_atom_name(data->target);
    
    /* Lookup the corresponding main clist row through the main clist
       object data. */
    target_row = (gint)gtk_object_get_data(GTK_OBJECT(clist_notebook),
        target_name)-1;
    
    /* Free the name of the target. */
    g_free(target_name);
    
    if(target_row >= 0)
    {
        /* Get copy of the previous target data that was captured when
           the buffer was created. */
        target_copy = gtk_clist_get_row_data(clist_notebook, target_row);
        
        /* Set the parameters in the selection data to mimic the captured
           selection exactly (except for the TIMESTAMP target). */
        gtk_selection_data_set(data, target_copy->type,
            target_copy->format, target_copy->data, target_copy->length);
    }
}

/************************************************************************
 * Signal handler that is called when the ownership of the selection    *
 * has been taken by another client. Its check however to see if it     *
 * wasn't a client of this utility before changing the owner button     *
 * and retrieving the data of the new selection.                        *
 ************************************************************************/
void widget_selection_clear(GtkWidget *widget,
         GdkEventSelection *event, GtkWindow *window)
{
    if(!is_owner_gclipper(window, event->selection) &&
       XGetSelectionOwner(gdk_display, event->selection))
    /* XGetSelectionOwner is an XLib function that has to be used to check
       if the selection is really not owned by a client.
       gdk_selection_owner_get() returns NULL if the selection is owned by
       a non-gtk (gdk) client. If the selection isn't owned by anyone, it
       shouldn't be re-fetched automatically. */
    {
        set_owner_button_off(window);
        selection_get(GTK_WIDGET(window), event->selection);
    }
}

/************************************************************************
 * Make a string representation of a string ;-) Now this is easy !      *
 ************************************************************************/
gchar *stringToString(gchar *source_string)
{
    gchar *string = NULL;
    
    string = g_strdup(source_string);
    
    return string;
}

/************************************************************************
 * Make a string representation of an integer.                          *
 ************************************************************************/
gchar *integerToString(gint integer)
{
    gchar *string = NULL;
    gchar buffer[20];
    
    sprintf(buffer, "%d", integer);
    string = g_strdup(buffer);

    return string;
}

/************************************************************************
 * Make a string representation of an unsigned integer.                 *
 ************************************************************************/
gchar *unsignedIntegerToString(guint integer)
{
    gchar *string = NULL;
    gchar buffer[20];
    
    sprintf(buffer, "%u", integer);
    string = g_strdup(buffer);

    return string;
}

/************************************************************************
 * Make a string representation of an array of atoms.                   *
 ************************************************************************/
gchar *atomsToString(GdkAtom *atoms, gint length)
{
    GString *string = NULL;
    guint   counter = 0;
    gchar   *atom_name = NULL;
    
    string = g_string_new("");
    
    /* Go over every atom in the array and append it to the string by
       using a linebreak as seperator. */
    for(counter = 0; counter < length / sizeof(GdkAtom);
        counter++)
    {
        /* Get the string representation of the atom. */
        atom_name = gdk_atom_name(atoms[counter]);
        if(atom_name != NULL)
        {
            /* Append the atom name to the string. */
            string = g_string_append(string, atom_name);
            string = g_string_append(string, "\n");
            g_free(atom_name);
        }
    }

    return string->str;
}

/************************************************************************
 * Display a label with the provided text on the right side of the      *
 * detailed info panel.                                                 *
 ************************************************************************/
GtkWidget* displayLabelString(GtkFrame *frame, gchar *string)
{
    GtkWidget *label = NULL;
    
    if(frame->bin.child)
    {
        gtk_widget_destroy(frame->bin.child);
    }

    label = gtk_label_new(string);
    
    return label;
}

/************************************************************************
 * Display a text widget with the provided text on the right side of    *
 * the detailed info panel.                                             *
 ************************************************************************/
GtkWidget* displayTextString(GtkFrame *frame, gchar *string)
{
    GtkObject *adjustment_text_vert = NULL;
    GtkWidget *scrolled_window = NULL;
    GtkWidget *text = NULL;

    if(frame->bin.child)
    {
        gtk_widget_destroy(frame->bin.child);
    }

    adjustment_text_vert = gtk_adjustment_new(0, 0, 0, 0, 0, 0);
    
    scrolled_window = gtk_scrolled_window_new(NULL,
        GTK_ADJUSTMENT(adjustment_text_vert));
    gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrolled_window),
        GTK_POLICY_NEVER, GTK_POLICY_AUTOMATIC);

    text = gtk_text_new(NULL, GTK_ADJUSTMENT(adjustment_text_vert));
    gtk_text_set_editable(GTK_TEXT(text), FALSE);
    gtk_text_set_word_wrap(GTK_TEXT(text), FALSE);
    gtk_text_set_line_wrap(GTK_TEXT(text), TRUE);
    gtk_container_add(GTK_CONTAINER(scrolled_window), text);
    
    gtk_text_insert(GTK_TEXT(text), NULL, NULL, NULL, string, -1);

    return scrolled_window;
}

/************************************************************************
 * Signal handler that is called when a row in the notebook clist has   *
 * been selected. This tries to make a visible representation of the    *
 * data that correponds to selection target defined by the clist row.   *
 ************************************************************************/
void clist_notebook_select_row(GtkCList *clist, gint row, gint column,
        GdkEventButton *event, GtkWidget *window)
{
    GtkWidget         *notebook = NULL;
    GtkWidget         *frame = NULL;
    GtkWidget         *hbox = NULL;
    GtkWidget         *page_data_widget = NULL;
    GtkWidget         *page_table = NULL;
    GtkWidget         *page_type_label = NULL;
    GtkWidget         *page_type_entry = NULL;
    GtkWidget         *page_format_label = NULL;
    GtkWidget         *page_format_entry = NULL;
    GtkWidget         *page_length_label = NULL;
    GtkWidget         *page_length_entry = NULL;
    struct TargetCopy *target_copy = NULL;
    gchar             *type_string = NULL;
    gchar             **text_list = NULL;
    gint              text_list_count = 0;
    gint              counter = 0;
    GString           *temp_gstring = NULL;
    gchar             *temp_text = NULL;
    
    notebook = gtk_object_get_data(GTK_OBJECT(window), "notebook");
    hbox = gtk_notebook_get_nth_page(GTK_NOTEBOOK(notebook),
        gtk_notebook_current_page(GTK_NOTEBOOK(notebook)));
    frame = gtk_object_get_data(GTK_OBJECT(hbox), "frame");
    
    /* Get the copy of the target structure from the row that has been
       clicked. */
    target_copy = (struct TargetCopy *)gtk_clist_get_row_data(clist, row);
    
    /* Create a representation for the targets that are supported
       or create a label that tells why it can't be displayed. */
    if(!target_copy || (guint)target_copy == NO_TARGETCOPY)
    {   
        page_data_widget = displayLabelString(GTK_FRAME(frame),
            "Unsupported target.");
    }
    else if(target_copy->data)
    {
        /* Get the string representation of the target type. */
        type_string = gdk_atom_name(target_copy->type);

        if(strcmp(type_string, "STRING") == 0)
        {
            temp_text = stringToString((gchar*)target_copy->data);
            page_data_widget = displayTextString(GTK_FRAME(frame),
                temp_text);
            g_free(temp_text);
        }
        else if(strcmp(type_string, "COMPOUND_TEXT") == 0)
        {
            temp_gstring = g_string_new("");
            
            /* Get the text list elements. */
            text_list_count =
                gdk_text_property_to_text_list(target_copy->type,
				    target_copy->format, target_copy->data,
					target_copy->length, &text_list);
                    
            /* Concanate the text list elements for display. */
            for(counter = 0; counter < text_list_count; counter++)
            {
                g_string_append(temp_gstring, text_list[counter]);
            }
                    
            page_data_widget = displayTextString(GTK_FRAME(frame),
                temp_gstring->str);

            if(text_list)
            {
                gdk_free_text_list(text_list);
            }
            
            g_string_free(temp_gstring, TRUE);
        }
        else if(strcmp(type_string, "INTEGER") == 0)
        {
            temp_text =
                unsignedIntegerToString(*((guint*)target_copy->data));
            page_data_widget = displayTextString(GTK_FRAME(frame),
                temp_text);
            g_free(temp_text);
        }
        else if(strcmp(type_string, "ATOM") == 0)
        {
            temp_text = atomsToString((GdkAtom *)target_copy->data,
                target_copy->length);
            page_data_widget = displayTextString(GTK_FRAME(frame),
                temp_text);
            g_free(temp_text);
        }
        else
        {
            page_data_widget = displayLabelString(GTK_FRAME(frame),
                "This target can't be displayed.");
        }
        
        g_free(type_string);
    }

    if(page_data_widget)
    {
        page_table = gtk_table_new(4, 2, FALSE);
        gtk_container_add(GTK_CONTAINER(frame), page_table);
        
        gtk_table_attach(GTK_TABLE(page_table), page_data_widget,
            0, 2, 0, 1,
            GTK_EXPAND|GTK_FILL|GTK_SHRINK,GTK_EXPAND|GTK_FILL|GTK_SHRINK,
            0, 0);

        if(target_copy && (guint)target_copy != NO_TARGETCOPY)
        {
            page_type_label = gtk_label_new(" Type ");
            gtk_table_attach(GTK_TABLE(page_table), page_type_label,
                0, 1, 1, 2,
                GTK_FILL, 0,
                0, 0);
        
            page_type_entry = gtk_entry_new();
            gtk_entry_set_editable(GTK_ENTRY(page_type_entry), FALSE);
            temp_text = gdk_atom_name(target_copy->type);
            if(temp_text)
            {
                gtk_entry_set_text(GTK_ENTRY(page_type_entry), temp_text);
                g_free(temp_text);
            }
            gtk_table_attach(GTK_TABLE(page_table), page_type_entry,
                1, 2, 1, 2,
                GTK_EXPAND|GTK_FILL|GTK_SHRINK, 0,
                0, 0);

            page_format_label = gtk_label_new(" Format ");
            gtk_table_attach(GTK_TABLE(page_table), page_format_label,
                0, 1, 2, 3,
                GTK_FILL, 0,
                0, 0);
        
            page_format_entry = gtk_entry_new();
            gtk_entry_set_editable(GTK_ENTRY(page_format_entry), FALSE);
            temp_text = integerToString(target_copy->format);
            gtk_entry_set_text(GTK_ENTRY(page_format_entry), temp_text);
            g_free(temp_text);
            gtk_table_attach(GTK_TABLE(page_table), page_format_entry,
                1, 2, 2, 3,
                GTK_EXPAND|GTK_FILL|GTK_SHRINK, 0,
                0, 0);

            page_length_label = gtk_label_new(" Length ");
            gtk_table_attach(GTK_TABLE(page_table), page_length_label,
                0, 1, 3, 4,
                GTK_FILL, 0,
                0, 0);
        
            page_length_entry = gtk_entry_new();
            gtk_entry_set_editable(GTK_ENTRY(page_length_entry), FALSE);
            temp_text = integerToString(target_copy->length);
            gtk_entry_set_text(GTK_ENTRY(page_length_entry), temp_text);
            g_free(temp_text);
            gtk_table_attach(GTK_TABLE(page_table), page_length_entry,
                1, 2, 3, 4,
                GTK_EXPAND|GTK_FILL|GTK_SHRINK, 0,
                0, 0);
        }
        
        gtk_widget_show_all(page_table);
    }
}

/************************************************************************
 * Free the data that is attached to every notebook clist row.          *
 * This a TargetCopy structure and the eventual data element allocated  *
 * for it.                                                              *
 ************************************************************************/
void clist_notebook_free_all_rows_data(GtkCList *clist)
{
    guint    counter = 0;
    gpointer row_data = NULL;
    gpointer target_name = NULL;
    gpointer target_data = NULL;
    
    /* Iterate over all the clist rows. */
    for(counter = 0; counter < clist->rows; counter++)
    {
        /* Get the clist row data. */
        row_data = gtk_clist_get_row_data(clist, counter);
        if(row_data && (guint)row_data != NO_TARGETCOPY)
        {
            /* Get the target name and free it if it exists. */
            target_name = ((struct TargetCopy *)row_data)->name;
            if(target_name)
            {
                g_free(target_name);
            }
            /* Get the target data and free it if it exists. */
            target_data = ((struct TargetCopy *)row_data)->data;
            if(target_data)
            {
                g_free(target_data);
            }
            g_free(row_data);
        }
    }
}

/************************************************************************
 * Signal handler that is called when the clist widget is destroyed.    *
 * This ensures that all additionally allocated data get freed.         *
 ************************************************************************/
gboolean clist_notebook_destroy(GtkCList *clist, GdkEvent *event,
             GtkWindow *window)
{
    clist_notebook_free_all_rows_data(clist);
    
    return FALSE;
}

/************************************************************************
 * Remove the notebook page widget with the provided index. The page    *
 * clist is destroyed also, thus freeing all the attached data.         *
 ************************************************************************/
void notebook_destroy_page(GtkNotebook *notebook,
         gint current_notebook_page)
{
    GtkWidget *notebook_page = NULL;
    GtkWidget *clist = NULL;
    
    /* Get the notebook page widget. */
    notebook_page = gtk_notebook_get_nth_page(notebook,
        current_notebook_page);

    /* Destroy the clist widget. */
    clist = gtk_object_get_data(GTK_OBJECT(notebook_page), "clist");
    gtk_widget_destroy(clist);
    
    /* Remove the notebook page. */
    gtk_notebook_remove_page(notebook, current_notebook_page);
}

/************************************************************************
 * Signal handler that will be called when the notebook widget is being *
 * destroyed. It ensures that all the attached bufferes and their data  *
 * are destroyed too.                                                   *
 ************************************************************************/
gboolean notebook_destroy(GtkNotebook *notebook, GdkEvent *event,
             GtkWindow *window)
{
    gint current_notebook_page = 0;
    
    /* Go over all the notebook pages and destroy them. */
    while(g_list_length(notebook->children))
    {
        current_notebook_page = gtk_notebook_current_page(notebook);
        notebook_destroy_page(notebook, current_notebook_page);
    }

    return FALSE;
}

/************************************************************************
 * Remove the selection buffer with the provided identifier.            *
 ************************************************************************/
void selection_remove(GtkWindow *window, guint selection_id)
{
    GtkWidget *clist_main = NULL;
    GtkWidget *notebook = NULL;
    guint     column_width = 0;

    clist_main = gtk_object_get_data(GTK_OBJECT(window), "clist");
    notebook = gtk_object_get_data(GTK_OBJECT(window), "notebook");

    gtk_clist_remove(GTK_CLIST(clist_main), selection_id);
    column_width =
        gtk_clist_optimal_column_width(GTK_CLIST(clist_main), 0); 
    gtk_clist_set_column_width(GTK_CLIST(clist_main), 0,
        column_width);
    notebook_destroy_page(GTK_NOTEBOOK(notebook), selection_id);

    synchronize_state(window);
    
    /* If there are no other buffers present, the main window will become
       the new primary selection owner. Otherwise, the next clist row will
       automatically be selected, making the clist of the corresponding
       notebook page the new owner. */
    if(!GTK_CLIST(clist_main)->rows)
    {
        become_primary_selection_owner(GTK_WINDOW(window),
            GTK_WIDGET(window));
    }
    else
    {
        /* If the deleted row was the last one in the main clist, make the
           previous one the newly selected row. */
        if(selection_id > GTK_CLIST(clist_main)->rows - 1)
        {
            selection_id = GTK_CLIST(clist_main)->rows - 1;
        }

        gtk_clist_select_row(GTK_CLIST(clist_main), selection_id, 0);
    }
}

/************************************************************************
 * Signal handler that is called when the delete button is clicked.     *
 ************************************************************************/
void button_delete_button_release(GtkEventBox *button,
         GdkEventButton *event, GtkWindow *window)
{
    GtkWidget *clist_main = NULL;
    gint      selected_row = 0;
    
    clist_main = gtk_object_get_data(GTK_OBJECT(window), "clist");
    
    /* If the shift key is pressed at the same time as the delete button,
       are all buffers. */
    if(event->state & GDK_SHIFT_MASK)
    {
        while(GTK_CLIST(clist_main)->rows)
        {
            selection_remove(window, 0);
        }
    }
    /* Otherwise erase just the selected one. */
    else
    {
        selected_row = (gint)GTK_CLIST(clist_main)->selection->data;
    
        selection_remove(window, selected_row);
    }
}

/************************************************************************
 * Signal handler that is called when the details button is toggled.    *
 ************************************************************************/
void button_details_toggled(GtkToggleButton *button, GtkWindow *window)
{
    GtkWidget *window_notebook = NULL;

    window_notebook = gtk_object_get_data(GTK_OBJECT(window),
        "window_notebook");
    
    if(gtk_toggle_button_get_active(button))
    {
        gtk_widget_show_all(window_notebook);
    }
    else
    {
        gtk_widget_hide(window_notebook);
    }
}

/************************************************************************
 * Signal handler that is called when the mousebutton is pressed on the *
 * owner button.                                                        *
 ************************************************************************/
void button_owner_button_press(GtkEventBox *button,
         GdkEventButton *event, GtkWindow *window)
{
    /* Stop the emission of this signal to prevent the normal widget
       toggle switch to take place. */
	gtk_signal_emit_stop_by_name(GTK_OBJECT(button),
        "button_press_event");
}

/************************************************************************
 * Signal handler that is called when the mousebutton is release on the *
 * owner button. This allows users to explicitely set GClipper's        *
 * auto-fetch mode on or off.                                           *
 ************************************************************************/
void button_owner_button_release(GtkEventBox *button,
         GdkEventButton *event, GtkWindow *window)
{
    GtkWidget *clist_main = NULL;
    GtkWidget *notebook = NULL;
    GtkWidget *notebook_page = NULL;
    GtkWidget *clist_notebook = NULL;

    /* Stop the emission of this signal to prevent the normal widget
       toggle switch to take place. */
	gtk_signal_emit_stop_by_name(GTK_OBJECT(button),
        "button_release_event");

    clist_main = gtk_object_get_data(GTK_OBJECT(window), "clist");
    
    /* If the selection is owned by GClipper, it should be given up. This
       means turning the status light from green to red. */
    if(is_owner_gclipper(window, SELECTION_PRIMARY))
    {
        set_owner_button_off(window);
        
        give_up_selection_ownership(window, SELECTION_PRIMARY);
        give_up_selection_ownership(window, SELECTION_CLIPBOARD);
    }
    /* If the selection is owned by another client, GClipper has to become
       the new owner, thus turning the status light from red to green. */
    else
    {
        set_owner_button_on(window);

        /* If there are rows (and thus selection buffers) present, set
           the first buffer as the owner of the primary and clipboard
           selection. */
        if(GTK_CLIST(clist_main)->rows)
        {
            notebook = gtk_object_get_data(GTK_OBJECT(window),
                                           "notebook");
            notebook_page =
                gtk_notebook_get_nth_page(GTK_NOTEBOOK(notebook), 0);
            clist_notebook =
                gtk_object_get_data(GTK_OBJECT(notebook_page), "clist");
            
            become_primary_selection_owner(GTK_WINDOW(window),
                clist_notebook);
            become_clipboard_selection_owner(GTK_WINDOW(window),
                clist_notebook);
        }
        /* If the buffer list is empty, the main window will become the
           primary and clipboard selection owner. */
        else
        {
            become_primary_selection_owner(window, GTK_WIDGET(window));
            become_clipboard_selection_owner(window, GTK_WIDGET(window));
        }
    }
}

/************************************************************************
 * Add a new notebook page to the notebook.                             *
 ************************************************************************/
GtkWidget *notebook_page_add(GtkWindow *window)
{
    GtkObject *adjustment_clist_vert = NULL;
    GtkWidget *clist = NULL;
    GtkWidget *frame = NULL;
    GtkWidget *hbox_clist = NULL;
    GtkWidget *hbox_page = NULL;
    GtkWidget *notebook = NULL;
    GtkWidget *vscrollbar = NULL;
    GList     *clist_notebook_windows = NULL;
    
    notebook = gtk_object_get_data(GTK_OBJECT(window), "notebook");
    clist_notebook_windows = gtk_object_get_data(GTK_OBJECT(window),
        "clist_notebook_windows");

    hbox_page = gtk_hbox_new(FALSE, 0);
    
    adjustment_clist_vert = gtk_adjustment_new(0, 0, 0, 0, 0, 0);
    
    hbox_clist = gtk_hbox_new(FALSE, 5);
    gtk_box_pack_start(GTK_BOX(hbox_page), hbox_clist, FALSE, FALSE, 0);
    
    clist = gtk_clist_new(1);
    gtk_clist_set_column_title(GTK_CLIST(clist), 0, "TARGETS");
    gtk_clist_set_selection_mode(GTK_CLIST(clist), GTK_SELECTION_BROWSE);
    gtk_clist_column_titles_show(GTK_CLIST(clist));
    gtk_clist_column_titles_passive(GTK_CLIST(clist));
    gtk_clist_set_vadjustment(GTK_CLIST(clist),
        GTK_ADJUSTMENT(adjustment_clist_vert)),
    gtk_signal_connect(GTK_OBJECT(clist), "selection_received",
        (GtkSignalFunc)clist_notebook_selection_received, window);
    gtk_signal_connect_after(GTK_OBJECT(clist), "selection_clear_event",
        (GtkSignalFunc)widget_selection_clear, window);
    gtk_signal_connect(GTK_OBJECT(clist), "selection_get",
        (GtkSignalFunc)clist_notebook_selection_get, window);
    gtk_signal_connect(GTK_OBJECT(clist), "select_row",
        (GtkSignalFunc)clist_notebook_select_row, window);
    gtk_signal_connect(GTK_OBJECT(clist), "destroy",
        (GtkSignalFunc)clist_notebook_destroy, window);
    gtk_box_pack_start(GTK_BOX(hbox_clist), clist, TRUE, TRUE, 0);
    
    vscrollbar =
        gtk_vscrollbar_new(GTK_ADJUSTMENT(adjustment_clist_vert));
    gtk_box_pack_start(GTK_BOX(hbox_page), vscrollbar, FALSE, FALSE, 0);
    
    frame = gtk_frame_new(NULL);
    gtk_frame_set_shadow_type(GTK_FRAME(frame), GTK_SHADOW_NONE);
    gtk_box_pack_end(GTK_BOX(hbox_page), frame, TRUE, TRUE, 0);
    
    gtk_object_set_data(GTK_OBJECT(hbox_page), "clist", (gpointer)clist);
    gtk_object_set_data(GTK_OBJECT(hbox_page), "frame", (gpointer)frame);

    gtk_widget_show_all(hbox_page);

    gtk_notebook_prepend_page(GTK_NOTEBOOK(notebook), hbox_page, NULL);
    gtk_notebook_set_page(GTK_NOTEBOOK(notebook), 0);
    
    /* The clist widget is realized first to be sure that there's a
       GdkWindow present in it's structure. */
    gtk_widget_realize(clist);
    /* Add the GdkWindow of the notebook clist to the list of possible
       selection owners. This list will be used to check against in the
       is_owner_gclipper() function. */
    clist_notebook_windows = g_list_prepend(clist_notebook_windows,
        GTK_WIDGET(clist)->window);
    gtk_object_set_data(GTK_OBJECT(window), "clist_notebook_windows",
        clist_notebook_windows);
    
    return hbox_page;
}

/************************************************************************
 * Signal handler that is called when the main window is closed. Before *
 * destroying it's widgets, is should explicitely destroy the notebook  *
 * widget to ensure that all the allocated buffers and their data are   *
 * freed.                                                               *
 ************************************************************************/
gboolean window_delete(GtkWidget *widget, GdkEvent *event,
             GtkWindow *window)
{
    GtkWidget *notebook = NULL;
    
    notebook = gtk_object_get_data(GTK_OBJECT(window), "notebook");

    gtk_widget_destroy(notebook);

    gtk_exit(0);

    return FALSE;
}

/************************************************************************
 * Signal handler that is called when the main window receives the data *
 * that corresponds to a selection target that has been requested       *
 * earlier. This function is the main dispatcher for the functionalities*
 * of the GClipper utility.                                             *
 ************************************************************************/
void window_selection_received(GtkWidget *widget, GtkSelectionData *data,
         guint time, GtkWindow *window)
{
    /* Static hashtables to store the previous selection owner and
       timestamp value. This will be used to provide a suitable timeout
       for possible double click selections.
       The TIMESTAMP target contains the time when the selection ownership
       occures. Together with the pointer to the actual selection owner,
       this provides a unique way of identifying each different selection.
    */
    static GHashTable *previous_owners = NULL;
    static GHashTable *previous_timestamps = NULL;

    Window            previous_owner;
    guint             previous_timestamp;
    Window            new_owner;
    guint             new_timestamp = 0;
    GtkWidget         *notebook = NULL;
    GtkWidget         *notebook_page = NULL;
    GtkWidget         *clist_main = NULL;
    GtkWidget         *clist_notebook = NULL;
    gchar             *target_type = NULL;
    gchar             *target_name = NULL;
    GdkAtom           *atoms_targets = NULL;
    GList             *glist_targets = NULL;
    gboolean          manual_get = FALSE;
    gboolean          process_targets = TRUE;
    guint             counter = 0;
    gint              clist_notebook_row = 0;
    guint             number_of_targets = 0;
    guint             column_width = 0;
    struct TargetCopy *target_copy = NULL;
    gchar             *temp_string = NULL; 
    GList             *temp_glist = NULL;
    guint             row_data = 0;
    
    if(!previous_owners)
    {
        previous_owners = g_hash_table_new(g_direct_hash, g_direct_equal);
    }
    if(!previous_timestamps)
    {
        previous_timestamps = g_hash_table_new(g_direct_hash,
            g_direct_equal);
    }
    
    previous_owner = (GdkAtom)g_hash_table_lookup(previous_owners,
        (gpointer)data->selection);
    previous_timestamp = (guint)g_hash_table_lookup(previous_timestamps,
        (gpointer)data->selection);
    
    clist_main = gtk_object_get_data(GTK_OBJECT(window), "clist");
    notebook = gtk_object_get_data(GTK_OBJECT(window), "notebook");
    
    /* Get the string representation of the target and target type
       atoms. */
    target_name = gdk_atom_name(data->target);
    target_type = gdk_atom_name(data->type);
    
    /* Get the owner of the current selection. This is done with the
       XGetSelectionOwner() function to get the real owner window.
       gdk_selection_owner_get() would just return NULL if the selection
       is owned by a non-gtk (gdk) client. */
    new_owner = XGetSelectionOwner(gdk_display, data->selection);
    
    /* Check which target has been requested, if it's a TIMESTAMP target
       this means that the double click timeout still has to be checked
       for. If it's a TARGETS target, the complete contents of all the
       selection targets should be duplicated if possible. */
    if(strcmp(target_name, "TIMESTAMP") == 0 &&
       !is_owner_gclipper(window, data->selection))
    {
        g_free(target_name);

        /* The following checks determine if the TIMESTAMP target was
           supported by the current selection owner. Normally this should
           always be the case since every selection is required to support
           it, but you never know. If the TIMESTAMP target is not
           supported, immediately request the retrieval of the TARGETS
           target. */
        if(target_type && data->length >= 0 && data->data != NULL)
        {
            /* Get the new timestamp data. */
            new_timestamp = *((guint*)data->data);
            
            /* Check if the previous owner, current owner and previous
               timeout, current timeout values are the same. If the is the
               case, request the retrieval of the TARGETS target. */
            if(previous_owner == new_owner &&
               previous_timestamp == new_timestamp)
            {
                get_target(GTK_WIDGET(window), "TARGETS",
                    data->selection);
            }
            else
            {
                /* Store the current selection owner and timeout for later
                   comparison. */
                g_hash_table_insert(previous_owners,
                    (gpointer)data->selection, (gpointer)new_owner);
                g_hash_table_insert(previous_timestamps,
                    (gpointer)data->selection, (gpointer)new_timestamp);
                
                /* Launch the selection retrieval again with a delay. This
                   delay allows double click selections to be performed
                   correctly and it's not too long to make to feel of the
                   utility sluggish. */
                if(data->selection == SELECTION_PRIMARY)
                {
                    gtk_timeout_add(250,
                        (GtkFunction)selection_get_primary, window);
                }
                else
                {
                    gtk_timeout_add(250,
                        (GtkFunction)selection_get_clipboard, window);
                }
            }
        }
        else
        {
            get_target(GTK_WIDGET(window), "TARGETS", data->selection);
        }
    }
    /* Check if the requested target was TARGETS. */
    else if(strcmp(target_name, "TARGETS") == 0)
    {
        g_free(target_name);

        /* The following checks determine if the TARGETS target was
           supported by the current selection owner. Normally this should
           always be the case since every selection is required to support
           it, but you never know. The target type checks are there to be
           sure that the data provided by the TARGETS target will be
           interpreted correctly. Normally the ATOM type should suffice,
           but some applications (like Rxvt) use the TARGETS type. */
        if(target_type &&
           (strcmp(target_type, "ATOM") == 0 ||
            strcmp(target_type, "TARGETS") == 0) &&
           data->length >= 0 && data->data != NULL)
        {
            /* Check if the selection was manually requested. */
            if(gtk_object_get_data(GTK_OBJECT(window), "MANUAL_GET"))
            {
                manual_get = TRUE;
            }
    
            /* Interprete the data provided by the TARGETS target. This
               is an array of atoms. By dividing the length of the data by
               the size of an atom, we know how many atoms are present in
               the array. The targets listed in the data will be stored 
               sorted in a GList structure and will be afterwards
               processed. */
            atoms_targets = (GdkAtom *)data->data;
            for(counter = 0; counter < (data->length / sizeof(GdkAtom));
                counter++)
            {
                /* Get the string representation of the target atom. */
                target_name = gdk_atom_name(atoms_targets[counter]);
                
                /* If the target name could be retrieved, store it in the
                   GList by using strcmp() as the function to sort the
                   GList contents. */
                if(target_name != NULL)
                {
                    glist_targets = g_list_insert_sorted(glist_targets,
                        target_name, (GCompareFunc)strcmp);
                    /* Some applications like NEdit, Netscape Mail Editor,
                       Wordperfect 8 ... use the X selection mechanism to
                       provide their internal selection functionality. If
                       those selections would be fetched automatically and
                       the owner forcebly changed, features likes copy,
                       cut, move and delete wouldn't be useable anymore.
                       They also become the selection owner at the
                       beginning of the selecting action (e.g. mousebutton
                       pressed) and not when the selecting action finishes
                       (e.g. mousebutton released). This behaviour would
                       not provide the intended selection data to the
                       GClipper utility. Instead only the first character
                       would be fetched.
                       These applications can be identified by the
                       presence of the DELETE and INSERT_SELECTION
                       targets. Those targets show that the selection is
                       intended to also change the selection data within
                       the owning application additionally to the mere
                       provision of the selected data to other clients.
                       When those targets are detected, the previously
                       acquired targets are not processed and the
                       ownership of the primary selection will not be
                       automatically regained. However, the monitoring of
                       the CLIPBOARD selection remains which offers a
                       logical way to fetch the selection data when it's
                       being copied or cut by the applications that were
                       mentioned above (ctrl+c, ctrl+x or alt+c, alt+x).*/
                    if(!manual_get &&
                       data->selection == SELECTION_PRIMARY &&
                       (strcmp(target_name, "DELETE") == 0 ||
                        strcmp(target_name, "INSERT_SELECTION") == 0))
                    {
                        process_targets = FALSE;
                    }
                }
            }
            
            /* The targets should be processed. */
            if(process_targets)
            {
                /* Before processing the targets, add a new page to the
                   notebook, thus creating a new clist that will be able
                   to contain the target names and their corresponding
                   data. */
                notebook_page = notebook_page_add(window);
                clist_notebook =
                    gtk_object_get_data(GTK_OBJECT(notebook_page),
                    "clist");
                    
                /* Process the targets by going over them one by one. */
                number_of_targets = g_list_length(glist_targets);
                for(counter = 0; counter < number_of_targets; counter++)
                {
                    /* Get the target name and create a new row in the
                       notebook clist widget with it. */
                    target_name = (gchar*)g_list_nth_data(glist_targets,
                        counter);
                    clist_notebook_row =
                        gtk_clist_append(GTK_CLIST(clist_notebook),
                        &target_name);
                        
                    /* Store which row corresponds to the target name by
                       setting it as the object data of the notebook clist
                       widget. A stupid trick was used to avoid the 0
                       value and to be able to check for the existance of
                       the target. That's why 1 was added to the row
                       number. Care should be taken when this information
                       is retrieved. To get the correct data, 1 should be
                       substracted from the result. */
                    gtk_object_set_data(GTK_OBJECT(clist_notebook),
                        target_name, (gpointer)clist_notebook_row+1);
                        
                    /* Some targets aren't supported by the system for
                       various reasons. The DELETE and INSERT_SELECTION
                       targets modify the selection within the owning
                       application. This is not something we wanna do, we
                       just wanna provide the same data. The MULTIPLE and
                       INSERT_PROPERTY targets are not supported by Gtk.
                       If the target is not one of those, it's data will
                       be requested and sent to the notebook clist widget.
                       The clist_notebook_selection_received() signal
                       handler will process the provided data and set a
                       copy of the target data as the data of the
                       corresponding clist row.
                       If the target is not supported, the data of the
                       clist row will be set to NULL which will be
                       correctly interpreted by the
                       clist_notebook_select_row() signal handler. */
                    if(strcmp(target_name, "DELETE") == 0 ||
                       strcmp(target_name, "MULTIPLE") == 0 ||
                       strcmp(target_name, "INSERT_PROPERTY") == 0 ||
                       strcmp(target_name, "INSERT_SELECTION") == 0)
                    {
                        gtk_clist_set_row_data(GTK_CLIST(clist_notebook),
                            counter, (gpointer)NO_TARGETCOPY);
                    }
                    else
                    {
                        get_target(clist_notebook, target_name,
                            data->selection);
                    }
                }
                
                /* Wait until all the targets have been processed and
                   attached to the notebook clist. */
                do
                {
                    /* Let Gtk process all the pending events before
                       continuing. */
                    while(gtk_events_pending())
                    {
                        gtk_main_iteration();
                    }
                    row_data =
                        (guint)gtk_clist_get_row_data(
                            GTK_CLIST(clist_notebook),
                            number_of_targets-1);
                }
                while(!row_data);
                
                /* If there was a STRING target, it's data will be used as
                   to add a new row to the main clist. Otherwise a new row
                   will be added saying that it couldn't find a STRING
                   target. */
                if(gtk_object_get_data(GTK_OBJECT(clist_notebook),
                   "STRING"))
                {
                    /* Get the notebook clist row number that corresponds
                       to the STRING target. */
                    clist_notebook_row = (gint)
                    gtk_object_get_data(GTK_OBJECT(clist_notebook),
                        "STRING") - 1;
                        
                    /* Get the notebook clist data that is attached to
                       previously found row number. This data is the
                       earlier made copy of the target data. */
                    target_copy = (struct TargetCopy *)
                        gtk_clist_get_row_data(GTK_CLIST(clist_notebook),
                        clist_notebook_row);
                        
                    /* Use the data of the STRING target as text for the
                       new main clist row. */
                    if(target_copy &&
                       (guint)target_copy != NO_TARGETCOPY &&
                       target_copy->data)
                    {
                        gtk_clist_prepend(GTK_CLIST(clist_main),
                            (gchar **)&(target_copy->data));
                    }
                    else
                    {
                        /* Add a row to the main clist with the mention
                           that it couldn't display the STRING target for
                           the selection. */
                        temp_string =
                            g_strdup("* UNDISPLAYABLE STRING TARGET FOR "
                                     "SELECTION *");
                        gtk_clist_prepend(GTK_CLIST(clist_main),
                            &temp_string);
                        g_free(temp_string);
                    }
                    
                    // move to the newly added row
                    gtk_clist_moveto(GTK_CLIST(clist_main), 0, 0, 0, 0);
                }
                else
                {
                    /* Add a row to the main clist with the mention that
                       it couldn't find a STRING target for the
                       selection. */
                    temp_string =
                        g_strdup("* NO STRING TARGET FOR SELECTION *");
                    gtk_clist_prepend(GTK_CLIST(clist_main),
                        &temp_string);
                    g_free(temp_string);
                }
                
                /* Check if the maximum number of selections hasn't been
                   exceeded and delete the last one if this is the
                   case. */
                if(GTK_CLIST(clist_main)->rows > MAX_SELECTIONS)
                {
                    selection_remove(window,
                        GTK_CLIST(clist_main)->rows - 1);
                }
                
                /* Select the newly created main clist row, calculate it's
                   width and set this as the new column width. */
                gtk_clist_select_row(GTK_CLIST(clist_main), 0, 0);
                column_width =
                    gtk_clist_optimal_column_width(GTK_CLIST(clist_main),
                    0); 
                gtk_clist_set_column_width(GTK_CLIST(clist_main), 0,
                    column_width);
    
                /* Select the first row of the new notebook clist,
                   calculate it's width and set this as the new column
                   width. */
                column_width =
                    gtk_clist_optimal_column_width(
                        GTK_CLIST(clist_notebook), 0); 
                gtk_clist_set_column_width(GTK_CLIST(clist_notebook), 0,
                    column_width);
                gtk_clist_select_row(GTK_CLIST(clist_notebook), 0, 0);

                /* Synchronize the state of the button widgets according
                   to the new number of buffers (notebook pages). */
                synchronize_state(window);
                
                /* Synchronize the state of the button widgets according
                   to the new number of buffers (notebook pages). */
                if(!GTK_CLIST(clist_main)->rows)
                {
                    become_primary_selection_owner(GTK_WINDOW(window),
                        GTK_WIDGET(window));
                }
            }
            /* The targets should not be processed. */
            else
            {
                /* Free the GList that contains the targets and also free
                   the strings that contain the names of the targets. */
                while(g_list_length(glist_targets))
                {
                    temp_glist = g_list_first(glist_targets);
                    glist_targets = g_list_remove_link(glist_targets,
                        temp_glist);
                    g_free(temp_glist->data);
                    g_list_free_1(temp_glist);
                }

                /* Launch the selection retrieval again with a delay. This
                   relaunch insures that newly occuring selections whose
                   targets should be processed, are automatically
                   fetched. If this behaviour would not be implemented,
                   then GClipper would sometimes just sit there, not
                   fetching anything anymore, thus becoming useless. It's
                   not very user-friendly to require the user to manually
                   make GClipper claim the ownership of the primary
                   before it starts automatically fetching selections
                   again. */
                if(data->selection == SELECTION_PRIMARY)
                {
                    gtk_timeout_add(250,
                        (GtkFunction)selection_get_primary, window);
                }
            }
        }
        /* The TARGETS target is apparently not supported by the
           selection. */
        else
        {
            /* Launch the selection retrieval again with a delay. See
               above for more explanation. */
            if(data->selection == SELECTION_PRIMARY)
            {
                gtk_timeout_add(250,
                    (GtkFunction)selection_get_primary, window);
            }
        }
        /* Remove the possible 'manually selected' flag. */
        gtk_object_remove_data(GTK_OBJECT(window), "MANUAL_GET");
    }
    
    /* If it exists, free the type name. */
    if(target_type)
    {
        g_free(target_type);
    }
}

/************************************************************************
 * Signal handler that is called when the details window is closed.     *
 * Instead of destroying the window and it's contents, the details      *
 * button should just be toggled, thus hiding the window and setting    *
 * the correct state on the details toggle button.                      *
 ************************************************************************/
gboolean window_notebook_delete(GtkWidget *widget, GdkEvent *event,
             GtkWindow *window)
{
    GtkWidget *button_details = NULL;

    button_details = gtk_object_get_data(GTK_OBJECT(window),
        "button_details");
        
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(button_details),
        FALSE);
    
    return TRUE;
}

/************************************************************************
 * Construct the main GClipper interface and store the pointers to the  *
 * widgets that have to be accessed later by using the object data of   *
 * the main window.                                                     *
 ************************************************************************/
GtkWidget *construct_main_interface()
{
    GtkWidget   *button_delete = NULL;
    GtkWidget   *button_details = NULL;
    GtkWidget   *button_owner = NULL;
    GtkWidget   *clist = NULL;
    GtkWidget   *notebook = NULL;
    GtkWidget   *scrolled_window = NULL;
    GtkWidget   *hbox_main = NULL;
    GtkWidget   *vbox_buttons = NULL;
    GtkWidget   *window = NULL;
    GtkWidget   *window_notebook = NULL;
    GtkTooltips *tooltips = NULL;
    GList       *clist_notebook_windows = NULL;
    
    tooltips = gtk_tooltips_new();
    gtk_tooltips_enable(tooltips);

    window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
    gtk_window_set_policy(GTK_WINDOW(window), TRUE, TRUE, FALSE);
    gtk_window_set_title(GTK_WINDOW(window),
        "GClipper v1.1 - by gbevin@thunderstorms.org");
    gtk_container_set_border_width(GTK_CONTAINER(window), 2);
    gtk_signal_connect(GTK_OBJECT(window), "delete_event",
        (GtkSignalFunc)window_delete, window);
    gtk_signal_connect(GTK_OBJECT(window), "selection_received",
        (GtkSignalFunc)window_selection_received, window);
    gtk_signal_connect(GTK_OBJECT(window), "selection_clear_event",
        (GtkSignalFunc)widget_selection_clear, window);

    hbox_main = gtk_hbox_new(FALSE, 5);
    gtk_container_add(GTK_CONTAINER(window), hbox_main);
    
    scrolled_window = gtk_scrolled_window_new(NULL, NULL);
    gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrolled_window),
        GTK_POLICY_NEVER, GTK_POLICY_AUTOMATIC);
    gtk_box_pack_start(GTK_BOX(hbox_main), scrolled_window,
        TRUE, TRUE, 0);
    
    clist = gtk_clist_new(1);
    gtk_widget_set_usize(clist, 100, -1);
    gtk_clist_set_selection_mode(GTK_CLIST(clist), GTK_SELECTION_BROWSE);
    gtk_clist_column_titles_hide(GTK_CLIST(clist));
    gtk_signal_connect(GTK_OBJECT(clist), "select_row",
        (GtkSignalFunc)clist_main_select_row, window);
    gtk_signal_connect(GTK_OBJECT(clist), "button_press_event",
        (GtkSignalFunc)clist_main_button_press, window);
    gtk_container_add(GTK_CONTAINER(scrolled_window), clist);
    gtk_clist_set_column_width(GTK_CLIST(clist), 0, 0);

    vbox_buttons = gtk_vbox_new(FALSE, 2);
    gtk_box_pack_start(GTK_BOX(hbox_main), vbox_buttons,
        FALSE, FALSE, 0);
    
    button_delete = gtk_button_new_with_label(" Delete ");
    gtk_signal_connect(GTK_OBJECT(button_delete), "button_release_event",
        (GtkSignalFunc)button_delete_button_release, window);
    gtk_box_pack_start(GTK_BOX(vbox_buttons), button_delete,
        FALSE, FALSE, 0);
    
    button_details = gtk_toggle_button_new_with_label(" Details ");
    gtk_signal_connect(GTK_OBJECT(button_details), "toggled",
        (GtkSignalFunc)button_details_toggled, window);
    gtk_box_pack_start(GTK_BOX(vbox_buttons), button_details,
        FALSE, FALSE, 0);

    button_owner = gtk_toggle_button_new_with_label(" Owner ");
    gtk_signal_connect(GTK_OBJECT(button_owner), "button_press_event",
        (GtkSignalFunc)button_owner_button_press, window);
    gtk_signal_connect(GTK_OBJECT(button_owner), "button_release_event",
        (GtkSignalFunc)button_owner_button_release, window);
    gtk_box_pack_start(GTK_BOX(vbox_buttons), button_owner,
        FALSE, FALSE, 0);

    gtk_widget_show_all(window);

    window_notebook = gtk_window_new(GTK_WINDOW_TOPLEVEL);
    gtk_window_set_policy(GTK_WINDOW(window_notebook), TRUE, TRUE, FALSE);
    gtk_window_set_title(GTK_WINDOW(window_notebook),
        "GClipper v1.1 - Details");
    gtk_container_set_border_width(GTK_CONTAINER(window_notebook), 0);
    gtk_signal_connect(GTK_OBJECT(window_notebook), "delete_event",
        (GtkSignalFunc)window_notebook_delete, window);

    notebook = gtk_notebook_new();
    gtk_notebook_set_show_tabs(GTK_NOTEBOOK(notebook), FALSE);
    gtk_signal_connect(GTK_OBJECT(notebook), "destroy",
        (GtkSignalFunc)notebook_destroy, window);
    gtk_container_add(GTK_CONTAINER(window_notebook), notebook);
    gtk_widget_set_usize(notebook, -1, 200);
    
    clist_notebook_windows = g_list_prepend(clist_notebook_windows,
        GTK_WIDGET(window)->window);

    gtk_object_set_data(GTK_OBJECT(window), "clist",
        (gpointer)clist);
    gtk_object_set_data(GTK_OBJECT(window), "notebook",
        (gpointer)notebook);
    gtk_object_set_data(GTK_OBJECT(window), "window_notebook",
        (gpointer)window_notebook);
    gtk_object_set_data(GTK_OBJECT(window), "button_delete",
        (gpointer)button_delete);
    gtk_object_set_data(GTK_OBJECT(window), "button_details",
        (gpointer)button_details);
    gtk_object_set_data(GTK_OBJECT(window), "button_owner",
        (gpointer)button_owner);
    gtk_object_set_data(GTK_OBJECT(window), "clist_notebook_windows",
        (gpointer)clist_notebook_windows);
    gtk_object_set_data(GTK_OBJECT(window), "tooltips",
        (gpointer)tooltips);
    
    return window;
}

int main(int argc, char *argv[])
{
    GtkWidget *window = NULL;
    
    /* Initialize Gtk. */
    gtk_init(&argc, &argv);
    
    SELECTION_CLIPBOARD = gdk_atom_intern("CLIPBOARD", FALSE); 
    
    /* Construct the Gtk interface elements and setup the signal handlers.
       Manually get the primary selection and claim the ownership of the
       clipboard selection. */
    window = construct_main_interface();
    become_primary_selection_owner(GTK_WINDOW(window), window);
    become_clipboard_selection_owner(GTK_WINDOW(window), window);
    synchronize_state(GTK_WINDOW(window));
    
    /* Starts the Gtk main loop. */
    gtk_main();
    
    return 0;
}
